import logging as log  # for verbose output
from .BitBucketRepoSession import BitBucketRepoSession
from .GitHubRepoSession import GitHubRepoSession
from .GitLabRepoSession import GitLabRepoSession
from .LocalVersionSession import LocalVersionSession
from .MercurialRepoSession import MercurialRepoSession


class HolderFactory:
    HOLDERS = {
        'github': GitHubRepoSession,
        'gitlab': GitLabRepoSession,
        'bitbucket': BitBucketRepoSession,
        'hg': MercurialRepoSession,
        'local': LocalVersionSession
    }

    @staticmethod
    # go through subclasses in order to find the one that is hodling a given project
    # repo is either complete URL or a name allowing to identify a single project
    def get_instance_for_repo(repo):
        holder_class = None
        hostname = None
        known_repo = None
        if repo.startswith(('https://', 'http://')):
            url_parts = repo.split('/')
            hostname = url_parts[2]
            for k, sc in HolderFactory.HOLDERS.items():
                if len(url_parts) >= (3 + sc.REPO_URL_PROJECT_COMPONENTS):
                    repo = "/".join(url_parts[3:3 + sc.REPO_URL_PROJECT_COMPONENTS])
                known_repo = sc.is_with_url(repo)
                if known_repo:
                    holder_class = sc
                    log.info('Using {} adapter'.format(k))
                    break
                if sc.matches_default_hostnames(hostname):
                    holder_class = sc
                    break
        if not holder_class:
            # the default:
            holder_class = HolderFactory.HOLDERS['github']
        if known_repo:
            repo = known_repo['repo']
            # known repo tells us hosted domain of e.g. mercurical web
            if 'hostname' in known_repo:
                hostname = known_repo['hostname']
        holder = holder_class(repo, hostname)
        if known_repo and 'branches' in known_repo:
            holder.set_branches(known_repo['branches'])
        return holder



import logging as log  # for verbose output
import os
import re
import time

import feedparser
from dateutil import parser

from .ProjectHolder import ProjectHolder
from .utils import asset_does_not_belong_to_machine, ApiCredentialsError


class GitHubRepoSession(ProjectHolder):
    DEFAULT_HOSTNAME = 'github.com'
    DEFAULT_HOLDER = True

    def __init__(self, repo, hostname):
        super(GitHubRepoSession, self).__init__()
        self.api_token = os.getenv("GITHUB_API_TOKEN")
        self.hostname = hostname
        if not self.hostname:
            self.hostname = self.DEFAULT_HOSTNAME
        # Explicitly specify the API version that we want:
        self.headers.update({
            'Accept': 'application/vnd.github.v3+json'
        })
        if self.api_token:
            log.info('Using API token.')
            self.headers.update({'Authorization': "token {}".format(self.api_token)})
        if self.hostname != self.DEFAULT_HOSTNAME:
            self.api_base = 'https://{}/api/v3'.format(self.hostname)
        else:
            self.api_base = 'https://api.{}'.format(self.DEFAULT_HOSTNAME)
        if '/' not in repo:
            r = self.get(
                '{}/search/repositories?q={}+in:name'.format(self.api_base, repo))
            self.repo = r.json()['items'][0]['full_name']
        else:
            self.repo = repo
        self.rate_limited_count = 0

    def get_rate_limit_url(self):
        return '{}/rate_limit'.format(self.api_base)

    def get(self, url, **kwargs):
        r = super(GitHubRepoSession, self).get(url, **kwargs)
        if r.status_code == 401:
            if self.api_token:
                raise ApiCredentialsError('API request was denied despite using an API token. '
                                          'Missing scopes?')
            raise ApiCredentialsError('Denied API access. Please set GITHUB_API_TOKEN env var '
                                      'as per https://github.com/dvershinin/lastversion#tips')
        if r.status_code == 403:
            if 'X-RateLimit-Reset' in r.headers and 'X-RateLimit-Remaining' in r.headers:
                if self.rate_limited_count > 2:
                    raise ApiCredentialsError(
                        'API requests were denied after retrying {} times'.format(
                            self.rate_limited_count)
                    )
                remaining = int(r.headers['X-RateLimit-Remaining'])
                wait_for = int(r.headers['X-RateLimit-Reset']) - time.time()
                if not remaining:
                    # got 403, likely due to used quota
                    if wait_for < 300:
                        if wait_for < 0:
                            log.warning(
                                'Exceeded API quota. Repeating request because quota is about to '
                                'be reinstated'
                            )
                        else:
                            log.warning(
                                'Waiting {} seconds for API quota reinstatement'.format(wait_for)
                            )
                            time.sleep(wait_for)
                        self.rate_limited_count = self.rate_limited_count + 1
                        return self.get(url)
                    else:
                        raise ApiCredentialsError(
                            'Exceeded API rate limit after waiting: {}'.format(
                                r.json()['message'])
                        )
            return self.get(url)

        if r.status_code == 403 and url != self.get_rate_limit_url():
            self.rate_limited_count = 0
        return r

    def rate_limit(self):
        url = '{}/rate_limit'.format(self.api_base)
        return self.get(url)

    def repo_query(self, uri):
        url = '{}/repos/{}{}'.format(self.api_base, self.repo, uri)
        return self.get(url)

    def repo_license(self, tag):
        r = self.repo_query('/license?ref={}'.format(tag))
        if r.status_code == 200:
            return r.json()
        return None

    def repo_readme(self, tag):
        r = self.repo_query('/readme?ref={}'.format(tag))
        if r.status_code == 200:
            return r.json()
        return None

    def get_latest(self, pre_ok=False, major=None):
        """
        Gets latest release satisfying "prereleases are OK" or major/branch constraints
        Strives to fetch formal API release if it exists, because it has useful information
        like assets
        """
        # data of selected tag, always contains ['version', 'tag_name', 'tag_date', 'type'] will
        # be returned
        ret = None

        # always fetch /releases as this will allow us to quickly look up if a tag is marked as
        # pre-release

        # then always get *all* tags through pagination

        # if pre not ok, filter out tags to check

        # if major, filter out tags to check for major

        # we need to check all tags commit dates simply because the most recent wins
        # we don't check tags which:
        # * marked pre-release in releases endpoints
        # * has a beta-like, non-version tag name

        # releases.atom and tags.atom don't differ much except releases having more data
        # yes, releases.atom include non-formal releases which are just tags, so we are good
        # based on testing, edited old releases don't jump forward and stay behind so they are great
        # the only downside is they don't bear pre-release mark (unlike API), and limited data
        # we work around both so we are fine to to use them for speed!
        r = self.get('https://{}/{}/releases.atom'.format(self.hostname, self.repo))
        # API requests are varied by cookie, we don't want serializer for cache fail because of that
        self.cookies.clear()
        feed = feedparser.parse(r.text)
        # TODO choose topmost (most recent), but do a doble check whether the edit was released (
        #  tag api)
        for tag in feed.entries:
            # https://github.com/apache/incubator-pagespeed-ngx/releases/tag/v1.13.35.2-stable
            tag_name = tag['link'].split('/')[-1]
            version = self.sanitize_version(tag_name, pre_ok, major)
            if version:
                # we always want to return formal release if it exists, cause it has useful data
                # grab formal release via APi to check for pre-release mark
                r = self.repo_query('/releases/tags/{}'.format(tag_name))
                if r.status_code == 200:
                    formal_release = r.json()
                    if not pre_ok and formal_release['prerelease']:
                        log.info(
                            "Found formal release for this tag which is unwanted "
                            "pre-release: {}.".format(version))
                        continue
                    # use full release info
                    ret = formal_release
                    ret['tag_name'] = tag_name
                    ret['tag_date'] = parser.parse(tag['updated'])
                    ret['version'] = version
                    ret['type'] = 'feed'
                else:
                    ret = tag
                    ret['tag_name'] = tag_name
                    ret['tag_date'] = parser.parse(tag['updated'])
                    ret['version'] = version
                    ret['type'] = 'feed'
                    # remove keys which are non-jsonable
                    # TODO use those (pop returns them)
                    ret.pop('updated_parsed', None)
                    ret.pop('published_parsed', None)
                log.info("Selected version as current selection: {}.".format(version))
                break
        # we are good with release from feeds only without looking at the API
        # simply because feeds list stuff in order of recency
        if ret:
            return ret

        # only if we did not find desired stuff through feeds, we switch to using API :)
        # this may be required in cases
        # releases.atom has limited tags, and all those are beta / invalid / non-versions
        # likewise, we want an older branch (major), which is not there in releases.atom
        # due to limited nature of data inside it

        # releases/latest fetches only non-prerelease, non-draft, so it
        # should not be used for hunting down pre-releases assets
        if not pre_ok:
            # https://stackoverflow.com/questions/28060116/which-is-more-reliable-for-github-api-conditional-requests-etag-or-last-modifie/57309763?noredirect=1#comment101114702_57309763
            # ideally we disable ETag validation for this endpoint completely
            r = self.repo_query('/releases/latest')
            if r.status_code == 200:
                tag_name = r.json()['tag_name']
                version = self.sanitize_version(tag_name, pre_ok, major)
                if version:
                    log.info("Selected version as current selection: {}.".format(version))
                    ret = r.json()
                    ret['tag_name'] = tag_name
                    ret['tag_date'] = parser.parse(r.json()['published_at'])
                    ret['version'] = version
                    ret['type'] = 'releases-latest'
        else:
            r = self.repo_query('/releases')
            if r.status_code == 200:
                for release in r.json():
                    tag_name = release['tag_name']
                    version = self.sanitize_version(tag_name, pre_ok, major)
                    if not version:
                        continue
                    if not ret or version > ret['version']:
                        log.info("Selected version as current selection: {}.".format(version))
                        ret = release
                        ret['tag_name'] = tag_name
                        ret['tag_date'] = parser.parse(release['published_at'])
                        ret['version'] = version
                        ret['type'] = 'release'

        # formal release may not exist at all, or be "late/old" in case
        # actual release is only a simple tag so let's try /tags
        r = self.repo_query('/tags?per_page=100')
        if r.status_code == 200:
            for t in r.json():
                tag_name = t['name']
                version = self.sanitize_version(tag_name, pre_ok, major)
                if not version:
                    continue
                c = self.repo_query('/git/commits/{}'.format(t['commit']['sha']))
                d = c.json()['committer']['date']
                d = parser.parse(d)

                if not ret or version > ret['version'] or d > ret['tag_date']:
                    # rare case: if upstream filed formal pre-release that passes as stable
                    # version (tag is 1.2.3 instead of 1.2.3b) double check if pre-release
                    # TODO handle API failure here as it may result in "false positive"?
                    release_for_tag = None
                    if not pre_ok:
                        r = self.repo_query('/releases/tags/{}'.format(tag_name))
                        if r.status_code == 200:
                            # noinspection SpellCheckingInspection
                            if r.json()['prerelease']:
                                log.info(
                                    "Found formal release for this tag which is unwanted "
                                    "pre-release: {}.".format(version))
                                continue
                            else:
                                release_for_tag = r.json()

                    log.info("Setting version as current selection: {}.".format(version))
                    if release_for_tag:
                        ret = release_for_tag
                        ret['tag_name'] = tag_name
                        ret['tag_date'] = parser.parse(release_for_tag['published_at'])
                        ret['version'] = version
                        ret['type'] = 'release'
                    else:
                        ret = t
                        ret['tag_name'] = tag_name
                        ret['tag_date'] = d
                        ret['version'] = version
                        ret['type'] = 'tag'

        return ret

    def release_download_url(self, release, shorter=False):
        """ The following format will benefit from:
        1) not using API, so is not subject to its rate limits
        2) likely has been accessed by someone in CDN and thus faster
        3) provides more or less unique filenames once the stuff is downloaded
        See https://fedoraproject.org/wiki/Packaging:SourceURL#Git_Tags
        We use variation of this: it does not need a parsed version (thus works for --pre better)
        and it is not broken on fancy release tags like v1.2.3-stable
        https://github.com/OWNER/PROJECT/archive/%{gittag}/%{gittag}-%{version}.tar.gz
        """
        ext = 'zip' if os.name == 'nt' else 'tar.gz'
        tag = release['tag_name']
        url_format = 'https://{}/{}/archive/{}/{}-{}.{}'
        if shorter:
            url_format = 'https://{}/{}/archive/{}.{}'
            return url_format.format(self.hostname, self.repo, tag, ext)
        return url_format.format(self.hostname, self.repo, tag, self.repo.split('/')[1], tag, ext)

    def get_assets(self, release, short_urls, assets_filter=None):
        urls = []
        if 'assets' in release and release['assets']:
            for asset in release['assets']:
                if assets_filter:
                    if not re.search(assets_filter, asset['name']):
                        continue
                else:
                    if asset_does_not_belong_to_machine(asset['name']):
                        continue
                urls.append(asset['browser_download_url'])
        else:
            download_url = self.release_download_url(release, short_urls)
            if not assets_filter or re.search(assets_filter, download_url):
                urls.append(download_url)
        return urls



import os
import platform
import re
import sys

import requests
import tqdm


class ApiCredentialsError(Exception):
    """raise this when there's a lookup error for my app"""


# matches os.name to known extensions that are meant *mostly* to run on it, and not other os.name-s
osExtensions = {
    'nt': '.exe',
    'posix': ('.tgz', '.tar.gz')
}

# matches *start* of sys.platform value to words in asset name
pfMarkers = {
    'win': ['windows', 'win'],
    'linux': ['linux'],
    'darwin': ['osx', 'darwin'],
    'freebsd': ['freebsd', 'netbsd', 'openbsd']
}

# matches platform.dist() ('redhat', '8.0', 'Ootpa') to words in asset name
# TODO use https://pypi.org/project/distro/
distroMarkers = {
    'centos': ['centos'],
    'redhat': ['centos', 'redhat']
}

# this is all too simple for now
nonAmd64Markers = ['i386', 'i686', 'arm', 'arm64', '386', 'ppc64', 'armv7',
                   'mips64', 'ppc64', 'mips64le', 'ppc64le']


def asset_does_not_belong_to_machine(asset):
    """
    Checks whether a given asset name is likely unusable on this machine
    An asset belongs to machine as long as this returns False
    :param asset:
    :type asset: str
    :return:
    :rtype:
    """
    # replace underscore with dash so that our shiny word boundary regexes won't break
    asset = asset.replace('_', '-')
    # bail if asset's extension "belongs" to other OS-es (simple)
    for osName, osExts in osExtensions.items():
        if os.name != osName and asset.endswith(osExts):
            return True
    # bail if asset's platform words belong to other platforms
    for pf, pfWords in pfMarkers.items():
        if not sys.platform.startswith(pf):
            for pfWord in pfWords:
                r = re.compile(r'\b{}(\d+)?\b'.format(pfWord), flags=re.IGNORECASE)
                matches = r.search(asset)
                if matches:
                    return True
    # weed out non-64 bit stuff from x86_64 bit OS
    # caution: may be false positive with 32 bit Python on 64 bit OS
    if platform.machine() in ['x86_64', 'AMD64']:
        for nonAmd64Word in nonAmd64Markers:
            r = re.compile(r'\b{}\b'.format(nonAmd64Word), flags=re.IGNORECASE)
            matches = r.search(asset)
            if matches:
                return True
            r = re.compile(r'\barm\d+\b', flags=re.IGNORECASE)
            matches = r.search(asset)
            if matches:
                return True
    # TODO weed out non-matching distros
    return False


def download_file(url, local_filename=None):
    if local_filename is None:
        local_filename = url.split('/')[-1]
    # NOTE the stream=True parameter below
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        # content-length may be empty, default to 0
        file_size = int(r.headers.get('Content-Length', 0))
        bar_size = 1024
        # fetch 8 KB at a time
        chunk_size = 8192
        # how many bars are there in a chunk?
        chunk_bar_size = chunk_size / bar_size
        # bars are by KB
        num_bars = int(file_size / bar_size)

        pbar = tqdm.tqdm(
            disable=None,  # disable on non-TTY
            total=num_bars,
            unit='KB',
            desc='Downloading {}'.format(local_filename),
            leave=True  # progressbar stays
        )
        with open(local_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=chunk_size):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)
                    # we fetch 8 KB, so we update progress by +8x
                    pbar.update(chunk_bar_size)
        pbar.set_description('Downloaded {}'.format(local_filename))
        pbar.close()

    return local_filename



from dateutil import parser

from .ProjectHolder import ProjectHolder


class BitBucketRepoSession(ProjectHolder):
    DEFAULT_HOSTNAME = 'bitbucket.org'
    KNOWN_REPO_URLS = {
        'mmonit.com': {'repo': 'tildeslash/monit'}
    }

    def __init__(self, repo, hostname):
        super(BitBucketRepoSession, self).__init__()
        self.hostname = hostname
        if not self.hostname:
            self.hostname = self.DEFAULT_HOSTNAME
        self.repo = repo

    def get_latest(self, pre_ok=False, major=None):
        response = self.get("https://api.bitbucket.org/2.0/repositories/{}/downloads".format(
            self.repo))
        data = response.json()
        release = data['values'][0]
        version = self.sanitize_version(release['name'], pre_ok, major)
        release['version'] = version
        release['tag_name'] = release['name']
        release['tag_date'] = parser.parse(release['created_on'])
        return release



# special case, private use now
# nginx version is taken as version of stable (written by rpm check script)
# to /usr/local/share/builder/nginx-stable.ver
import logging as log  # for verbose output
from .ProjectHolder import ProjectHolder


class LocalVersionSession(ProjectHolder):
    DEFAULT_HOSTNAME = None

    def __init__(self, repo, hostname):
        super(LocalVersionSession, self).__init__()
        self.repo = repo
        self.hostname = hostname

    def get_latest(self, pre_ok=False, major=None):
        if pre_ok:
            log.info('--pre is not supported for local version sources')
        if not major:
            major = 'stable'
        ver_file = '/usr/local/share/builder/{}-{}.ver'.format(self.repo, major)
        with open(ver_file, 'r') as file:
            version = file.read().replace('\n', '')
            return {
                'version': version,
                'tag_name': version
            }



import logging as log  # for verbose output
import os

from dateutil import parser

from .ProjectHolder import ProjectHolder


class GitLabRepoSession(ProjectHolder):
    DEFAULT_HOSTNAME = 'gitlab.com'

    def __init__(self, repo, hostname):
        super(GitLabRepoSession, self).__init__()
        self.pa_token = os.getenv("GITLAB_PA_TOKEN")
        self.hostname = hostname
        if not self.hostname:
            self.hostname = self.DEFAULT_HOSTNAME
        if self.pa_token:
            log.info('Using Personal Access token.')
            self.headers.update({'Private-Token': "{}".format(self.pa_token)})
        self.api_base = 'https://{}/api/v4'.format(self.hostname)
        self.repo = repo
        self.repo_id = self.repo.replace('/', '%2F')

    def repo_query(self, uri):
        url = '{}/projects/{}/repository{}'.format(self.api_base, self.repo_id, uri)
        return self.get(url)

    def get_latest(self, pre_ok=False, major=None):
        ret = None

        # gitlab returns tags by updated in desc order, this is just what we want :)
        r = self.repo_query('/tags')
        if r.status_code == 200:
            for t in r.json():
                tag = t['name']
                version = self.sanitize_version(tag, pre_ok, major)
                if not version:
                    continue
                if not ret or version > ret['version']:
                    log.info("Setting version as current selection: {}.".format(version))
                    ret = t
                    ret['tag_name'] = tag
                    ret['tag_date'] = parser.parse(t['commit']['created_at'])
                    ret['version'] = version
                    ret['type'] = 'tag'
                    # stop on first tag, because gitlab is good (c)
                    break
        return ret

    def release_download_url(self, release, shorter=False):
        if shorter:
            log.info('Shorter URLs are not supported for GitLab yet')
        # https://gitlab.com/onedr0p/sonarr-episode-prune/-/archive/v3.0.0/sonarr-episode-prune-v3.0.0.tar.gz
        ext = 'zip' if os.name == 'nt' else 'tar.gz'
        tag = release['tag_name']
        url_format = 'https://{}/{}/-/archive/{}/{}-{}.{}'
        return url_format.format(self.hostname, self.repo, tag, self.repo.split('/')[1], tag, ext)

    def repo_license(self, tag):
        # TODO implement
        pass



#!/usr/bin/env python
"""
lastversion
==========
.. code:: shell
  $ lastversion wp-cli/wp-cli
  $ lastversion apache/incubator-pagespeed-ngx
"""

from setuptools import find_packages, setup
import os
import re

_version_re = re.compile(r"__version__\s=\s'(.*)'")

# require at least requests==2.6.1 due to cachecontrol's bug:
# https://github.com/ionrock/cachecontrol/issues/137
install_requires = ['requests>=2.6.1', 'packaging', 'cachecontrol', 'lockfile', 'appdirs',
                    'python-dateutil', 'feedparser', 'PyYAML', 'tqdm']
tests_requires = ["pytest>=4.4.0", "flake8", "pytest-xdist"]

with open("README.md", "r") as fh:
    long_description = fh.read()

base_dir = os.path.dirname(__file__)

with open(os.path.join(base_dir, "lastversion", "__about__.py"), 'r') as f:
    version = _version_re.search(f.read()).group(1)

setup(
    name="lastversion",
    version=version,
    author="Danila Vershinin",
    author_email="info@getpagespeed.com",
    url="https://github.com/dvershinin/lastversion",
    description="A CLI tool to fetch last GitHub release version",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["tests"]),
    zip_safe=False,
    license="BSD",
    install_requires=install_requires,
    extras_require={
        "tests": install_requires + tests_requires,
    },
    tests_require=tests_requires,
    include_package_data=True,
    entry_points={"console_scripts": ["lastversion = lastversion:main"]},
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Operating System :: OS Independent",
        "Topic :: Software Development",
    ],
)



import feedparser
import datetime

from .ProjectHolder import ProjectHolder


class MercurialRepoSession(ProjectHolder):
    REPO_URL_PROJECT_COMPONENTS = 1
    KNOWN_REPO_URLS = {
        'nginx.org': {
            'repo': 'nginx',
            'hostname': 'hg.nginx.org',
            'branches': {
                'stable': '\\.\\d?[02468]\\.',
                'mainline': '\\.\\d?[13579]\\.'
            }
        }
    }

    @classmethod
    def matches_default_hostnames(cls, hostname):
        return hostname.startswith('hg.')

    def __init__(self, repo, hostname):
        super(MercurialRepoSession, self).__init__()
        self.hostname = hostname
        self.repo = repo

    def get_latest(self, pre_ok=False, major=None):
        ret = None
        # to leverage cachecontrol, we fetch the feed using requests as usual
        # then feed the feed to feedparser as a raw string
        # e.g. https://hg.nginx.org/nginx/atom-tags
        # https://pythonhosted.org/feedparser/common-atom-elements.html
        r = self.get('https://{}/{}/atom-tags'.format(self.hostname, self.repo))
        feed = feedparser.parse(r.text)
        for tag in feed.entries:
            tag_name = tag['title']
            version = self.sanitize_version(tag_name, pre_ok, major)
            if not version:
                continue
            if not ret or version > ret['version']:
                ret = tag
                tag['tag_name'] = tag['title']
                tag['version'] = version
                # converting from struct
                tag['tag_date'] = datetime.datetime(*tag['published_parsed'][:6])
        return ret



"""
lastversion
==========
License: BSD, see LICENSE for more details.
"""

import argparse
import json
import logging as log  # for verbose output
import os
import re
import sys

import yaml
from appdirs import user_cache_dir
from cachecontrol import CacheControl
from cachecontrol.caches.file_cache import FileCache
# from cachecontrol.heuristics import ExpiresAfter
from packaging.version import Version, InvalidVersion

from .ProjectHolder import ProjectHolder
from .HolderFactory import HolderFactory
from .__about__ import __version__
from .utils import download_file, ApiCredentialsError


def latest(repo, output_format='version', pre_ok=False, assets_filter=False,
           short_urls=False, major=None):
    cache_dir = user_cache_dir("lastversion")
    log.info("Using cache directory: {}.".format(cache_dir))

    repo_data = {
        'module_of': None
    }
    if repo.endswith('.yml'):
        with open(repo) as fpi:
            repo_data = yaml.safe_load(fpi)
            if 'repo' in repo_data:
                if 'nginx-extras' in repo:
                    repo_data['module_of'] = 'nginx'
                name = os.path.splitext(os.path.basename(repo))[0]
                if repo_data['module_of']:
                    name = 'nginx-module-{}'.format(name)
                repo = repo_data['repo']
                repo_data['name'] = name

    # find the right hosting for this repo
    project_holder = HolderFactory.get_instance_for_repo(repo)

    # we are completely "offline" for 1 hour, not even making conditional requests
    # heuristic=ExpiresAfter(hours=1)   <- make configurable
    with CacheControl(project_holder, cache=FileCache(cache_dir)) as s:
        release = s.get_latest(pre_ok=pre_ok, major=major)
    s.close()

    # bail out, found nothing that looks like a release
    if not release:
        return None

    version = release['version']
    tag = release['tag_name']

    # return the release if we've reached far enough:
    if output_format == 'version':
        return version

    if output_format == 'json':
        release['version'] = str(version)
        release['tag_date'] = str(release['tag_date'])
        release['v_prefix'] = tag.startswith("v")
        release['spec_tag'] = tag.replace(
            str(version),
            '%{upstream_version}' if repo_data['module_of'] else '%{version}'
        )
        release['tag_name'] = tag
        if hasattr(s, 'repo_license'):
            release['license'] = s.repo_license(tag)
        if hasattr(s, 'repo_readme'):
            release['readme'] = s.repo_readme(tag)
        release.update(repo_data)
        return release

    if output_format == 'assets':
        return s.get_assets(release, short_urls, assets_filter)

    if output_format == 'source':
        return s.release_download_url(release, short_urls)

    return None


def check_version(value):
    """
    Argument parser helper for --newer-than (-gt) option
    :param value:
    :type value:
    :return:
    :rtype:
    """
    try:
        value = Version(value)
    except InvalidVersion:
        raise argparse.ArgumentTypeError("%s is an invalid version value" % value)
    return value


def parse_version(tag):
    h = ProjectHolder()
    return h.sanitize_version(tag, pre_ok=True)


def main():
    parser = argparse.ArgumentParser(description='Get latest release from GitHub.',
                                     prog='lastversion')
    parser.add_argument('action', nargs='?', default='get', help='Special action to run, '
                                                                 'e.g. download, test')
    parser.add_argument('repo', metavar='REPO',
                        help='GitHub repository in format owner/name')
    # affects what is considered last release
    parser.add_argument('--pre', dest='pre', action='store_true',
                        help='Include pre-releases in potential versions')
    parser.add_argument('--verbose', dest='verbose', action='store_true')
    # no --download = False, --download filename.tar, --download = None
    parser.add_argument('-d', '--download', dest='download', nargs='?', default=False, const=None)
    # how / which data of last release we want to present
    # assets will give download urls for assets if available and sources archive otherwise
    # sources will give download urls for sources always
    # json always includes "version", "tag_name" etc + whichever json data was
    # used to satisfy lastversion
    parser.add_argument('--format',
                        choices=['version', 'assets', 'source', 'json'],
                        help='Output format')
    parser.add_argument('--assets', dest='assets', action='store_true',
                        help='Returns assets download URLs for last release')
    parser.add_argument('--source', dest='source', action='store_true',
                        help='Returns only source URL for last release')
    parser.add_argument('--version', action='version',
                        version='%(prog)s {version}'.format(version=__version__))
    parser.add_argument('-gt', '--newer-than', type=check_version, metavar='VER',
                        help="Output only if last version is newer than given version")
    parser.add_argument('-b', '--major', metavar='MAJOR',
                        help="Only consider releases of specific major "
                             "version, e.g. 2.1.x")
    parser.add_argument('--filter', metavar='REGEX', help="Filters --assets result by a regular "
                                                          "expression")
    parser.add_argument('-su', '--shorter-urls', dest='shorter_urls', action='store_true',
                        help='A tiny bit shorter URLs produced')
    parser.set_defaults(validate=True, verbose=False, format='version',
                        pre=False, assets=False, newer_than=False, filter=False,
                        shorter_urls=False, major=None)
    args = parser.parse_args()

    if args.repo == "self":
        args.repo = "dvershinin/lastversion"

    if args.verbose:
        log.basicConfig(format="%(levelname)s: %(message)s", level=log.DEBUG)
        log.info("Verbose output.")
    else:
        log.basicConfig(format="%(levelname)s: %(message)s")

    if args.assets:
        args.format = 'assets'

    if args.source:
        args.format = 'source'

    if args.filter:
        args.filter = re.compile(args.filter)

    if args.action == 'test':
        print(parse_version(args.repo))
        # TODO dynamic exit status
        sys.exit(0)

    if args.action == 'install':
        # we can only install assets
        args.format = 'assets'

    # imply source download, unless --assets specified
    # --download is legacy flag to specify download action or name of desired download file
    # --download == None indicates download intent where filename is based on upstream
    if args.action == 'download':
        if args.download is False:
            args.download = None

    if args.download is not False:
        args.action = 'download'
        if args.format != 'assets':
            args.format = 'source'

    # other action are either getting release or doing something with release (extend get action)
    try:
        res = latest(args.repo, args.format, args.pre, args.filter,
                     args.shorter_urls, args.major)
    except ApiCredentialsError as error:
        sys.stderr.write(str(error) + os.linesep)
        sys.exit(4)

    if res:
        # download command
        if args.action == 'download':
            if args.format == 'source':
                # there is only one source, but we need an array
                res = [res]
            for url in res:
                log.info("Downloading {} ...".format(url))
                # there can be many assets, so we do not "rename" them
                # there can be only one source, so we allow passing custom filename for it
                download_file(url, args.download if args.format == 'source' else None)
            sys.exit(0)

        if args.action == 'install':
            rpms = [asset for asset in res if asset.endswith('.rpm')]
            # if an installator is capable of fetching the URL, pass it directly
            # otherwise download to temp directory
            yum = "/usr/bin/yum"
            # TODO find out why dnf is downloading much slower than lastversion :-\
            if os.path.exists(yum):
                if not rpms:
                    log.error('No assets found to install')
                    sys.exit(1)
                for rpm in rpms:
                    import subprocess
                    subprocess.call(["/bin/sudo", yum, "install", rpm])
                    sys.exit(0)

        # display version in various formats:
        if args.format == 'assets':
            print("\n".join(res))
        elif args.format == 'json':
            json.dump(res, sys.stdout)
        else:
            print(res)
            # special exit code "2" is useful for scripting to detect if no newer release exists
            if args.newer_than:
                if res <= args.newer_than:
                    sys.exit(2)
    else:
        # empty list returned to --assets, emit 3
        if args.format == 'assets' and res is not False:
            sys.exit(3)
        sys.stderr.write("No release was found" + os.linesep)
        sys.exit(1)


if __name__ == "__main__":
    main()



import logging as log
import re
from .__about__ import __version__

import requests

# this class basically corresponds to something (often a website) which holds
# projects (usually a bunch). often this is a github-like website, so we subclass session
# but this also maybe something special, which either way can be used as a source of version
# information for a project based on its URL or name (see LocalVersionSession)
# it is instantantiated with a particular project in mind/set, but also has some methods for
# stuff like searching one
from packaging.version import InvalidVersion, Version


class ProjectHolder(requests.Session):
    # web accessible project holders may have single well-known domain usable by everyone
    # in case of GitHub, that is github.com, for Mercurial web gui - here isn't one, etc.
    DEFAULT_HOSTNAME = None
    DEFAULT_HOLDER = False
    KNOWN_REPO_URLS = {}
    # e.g. owner/project, but mercurial just /project together with hostname
    # adapter array should list how many elements make up "repo", e.g. for hg.nginx.com/repo it
    # is only one instead of 2
    # or a "format" specifier for matching
    REPO_URL_PROJECT_COMPONENTS = 2

    def __init__(self):
        super(ProjectHolder, self).__init__()
        self.headers.update({'User-Agent': 'lastversion/{}'.format(__version__)})
        log.info('Using {} project holder'.format(type(self).__name__))
        self.branches = None

    def set_branches(self, branches):
        self.branches = branches

    @classmethod
    def is_with_url(cls, repo):
        for url in cls.KNOWN_REPO_URLS:
            if repo.startswith((url, "https://{}".format(url), "http://{}".format(url))):
                log.info('{} Starts with {}'.format(repo, url))
                return cls.KNOWN_REPO_URLS[url]
        return False

    @classmethod
    def matches_default_hostnames(cls, hostname):
        return hostname == cls.DEFAULT_HOSTNAME

    def matches_major_filter(self, version, major):
        if self.branches and major in self.branches:
            if re.search(r"{}".format(self.branches[major]), version):
                log.info('{} matches major {}'.format(version, self.branches[major]))
                return True
        elif '{}.'.format(major) in version:
            log.info('{} is not under the desired major {}'.format(
                version, major))
            return True
        return False

    def sanitize_version(self, version, pre_ok=False, major=None):
        """extract version from tag name"""
        log.info("Checking tag {} as version.".format(version))
        if major and not self.matches_major_filter(version, major):
            log.info('{} is not under the desired major {}'.format(
                version, major))
            return False
        # many times they would tag foo-1.2.3 which would parse to LegacyVersion
        # we can avoid this, by reassigning to what comes after the dash:
        parts = version.split('-', 1)
        if len(parts) == 2 and parts[0].isalpha():
            version = parts[1]
        # help devel releases to be correctly identified
        # https://www.python.org/dev/peps/pep-0440/#developmental-releases
        version = re.sub('-devel$', '.dev0', version)
        # release-3_0_2 is often seen on Mercurial holders
        # note that above code removes "release-" already so we are left with "3_0_2"
        if re.search(r'^(?:\d+_)+(?:\d+)', version):
            version = version.replace('_', '.')
        try:
            v = Version(version)
            if not v.is_prerelease or pre_ok:
                log.info("Parsed as Version OK")
                log.info("String representation of version is {}.".format(v))
                return v
            log.info("Parsed as unwanted pre-release version: {}.".format(v))
            return False
        except InvalidVersion:
            log.info("Failed to parse {} as Version.".format(version))
            # attempt to remove extraneous chars and revalidate
            s = re.search(r'([0-9]+([.][0-9x]+)+(rc[0-9]?)?)', version)
            if s:
                version = s.group(1)
                log.info("Sanitized tag name value to {}.".format(version))
                if version.endswith('.x'):
                    # 1.10.x is a dev release without clear version, so even pre ok will not get it
                    return False
                # we know regex is valid version format, so no need to try catch
                return Version(version)
            log.info("Did not find anything that looks like a version in the tag")
            return False
